# -*- coding: utf-8 -*-

#  PycaWM
#  miniiconbox.py
#  Copyright (c) 2007 Vincent Rasneur, Anaël Verrier

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 only.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys

if sys.version_info < (2, 5):
    raise ImportError('miniiconbox module needs at least Python 2.5!')

import Image

from Xlib import X, display

from pycawm.basedirspec import load_first_data
from pycawm.hookmanager import add_post_hook, remove_post_hook
from pycawm.wrappers import DisplayWrapper
from pycawm import Client

class IconifiedClient(object):
    event_mask = (X.ExposureMask | X.ButtonPressMask |
                  X.EnterWindowMask | X.LeaveWindowMask)

    def __init__(self, box, client, x, y):
        self.client = client
        self.box = box
        self.name = str(client.name) if client.name else '(No title)'
        self.mini_icon_width = self.box.side
        self.mini_icon_height = self.box.side
        self.image = client.get_mini_icon((self.mini_icon_width,
                                           self.mini_icon_height))
        if self.image is None:
            if self.box.image_default is not None:
                self.image = self.box.image_default
            else:
                self.image = Image.new('RGB', (self.mini_icon_width,
                                               self.mini_icon_height))
        self.selected = False
        self.font = self.box.display.open_font('fixed')
        self.extents = self.font.query_text_extents(self.name)
        self.title_width = self.mini_icon_width
        self.title_height = self.extents.overall_ascent + 4
        root = self.box.display.screen().root
        self.mini_icon = root.create_window(
            x, y,
            self.mini_icon_width, self.mini_icon_height,
            1, self.box.wm.screen.root_depth,
            event_mask=self.event_mask,
            override_redirect=True)
        self.title = root.create_window(
            x, y + self.mini_icon_height + 4,
            self.title_width, self.title_height,
            1, self.box.wm.screen.root_depth,
            event_mask=self.event_mask,
            override_redirect=True)
        self.gc = self.mini_icon.create_gc(
            foreground=box.wm.get_color('white'),
            background=box.wm.get_color('#707070'),
            font=self.font)
        self.mini_icon.configure(stack_mode=X.Below)
        self.title.configure(stack_mode=X.Below)
        self.mini_icon.map()
        self.title.map()
        self.box.display.sync()

    def update(self):
        get_color = self.box.wm.get_color
        color = get_color('#606060' if self.selected else '#707070')
        self.mini_icon.put_pil_image(self.gc, 0, 0, self.image)
        self.gc.change(foreground=color)
        self.title.fill_rectangle(self.gc,
                                  0, 0,
                                  self.title_width, self.title_height)
        self.gc.change(foreground=get_color('white'),
                       background=color)
        self.title.image_text(self.gc, 0, self.title_height - 2, str(self.name))

    def contain(self, point_x, point_y):
        def contain_aux(geometry):
            return (point_x >= geometry.x and
                    point_x < (geometry.x + geometry.width + 2) and
                    point_y >= geometry.y and
                    point_y < (geometry.y + geometry.height + 2))
        geo_mini_icon = self.mini_icon.get_geometry()
        geo_title = self.title.get_geometry()
        return contain_aux(geo_mini_icon) or contain_aux(geo_title)

    def replace(self, x, y):
        # do not generate enternotify/leavenotify events!
        self.mini_icon.change_attributes(event_mask=X.NoEventMask)
        self.title.change_attributes(event_mask=X.NoEventMask)
        self.mini_icon.configure(x=x, y=y)
        self.title.configure(x=x, y=y + self.mini_icon_height + 4)
        # restore the normal windows event masks
        self.mini_icon.change_attributes(event_mask=self.event_mask)
        self.title.change_attributes(event_mask=self.event_mask)

    def enter(self):
        geo = self.title.get_geometry()
        title_width = self.extents.overall_width
        if title_width > self.title_width:
            self.title_width = title_width
        title_x = geo.x - (self.title_width - geo.width) / 2
        if title_x < 0:
            title_x = 0
        if title_x + self.title_width > self.box.wm.screen.width_in_pixels:
            title_x = self.box.wm.screen.width_in_pixels - self.title_width
        self.title.configure(x=title_x,
                             width=self.title_width,
                             stack_mode=X.Above)
        self.mini_icon.configure(stack_mode=X.Above)
        self.selected = True
        self.update()

    def leave(self):
        geo = self.mini_icon.get_geometry()
        self.title_width = self.mini_icon_width
        self.title.configure(x=geo.x,
                             width=self.title_width,
                             stack_mode=X.Below)
        self.mini_icon.configure(stack_mode=X.Below)
        self.selected = False
        self.update()

    def uniconify(self):
        self.client.uniconify()

    def destroy(self):
        self.mini_icon.unmap()
        self.mini_icon.destroy()
        self.title.unmap()
        self.title.destroy()
        self.gc.free()
        self.font.close()

class MiniIconBox(object):
    def __init__(self, wm, x=0, y=0, side=None, columns=3, image_default=None,
                 hide_transients=True):
        self.wm = wm
        self.display = display.Display(self.wm.display.get_display_name())
        self.display_wrapper = DisplayWrapper(self.display, self.handle_events)
        self.wm.displays.append(self.display_wrapper)
        self.iconified_clients = list()
        self.x = x
        self.y = y
        self.columns = columns
        if side is None:
            self.side = Client.title_bar_height
        else:
            self.side = side
        self.space = 8
        if image_default is not None:
            try:
                self.image_default = Image.open(
                    load_first_data(image_default)).resize((self.side,
                                                            self.side),
                                                           Image.ANTIALIAS)
            except IOError:
                self.image_default = None
        else:
            self.image_default = None
        self.hide_transients = hide_transients
        add_post_hook(Client.iconify, self.add_iconified_client)
        add_post_hook(Client.uniconify, self.remove_iconified_client)
        add_post_hook(Client.withdraw, self.withdraw_iconified_client)
        add_post_hook(Client.destroy, self.remove_iconified_client)
        self.add_existing_iconified_clients()

    def add_existing_iconified_clients(self):
        for client in self.wm.clients:
            if client.states.icccm_state == client.iconified:
                self.add_iconified_client(client)

    def add_iconified_client(self, client):
        # already iconified?
        if self.get_iconified_client_by_client(client) is not None:
            return

        if self.hide_transients:
            # main window for the transient already iconified?
            if self.get_iconified_client_by_client(
                self.wm.get_managed_client(client.transient_for)) is not None:
                return

            # remove all the previous transient clients
            transients = client.transient_clients
            for iconified_client in self.iconified_clients[:]:
                if iconified_client.client in transients:
                    self.remove_iconified_client(iconified_client.client)

        nb = len(self.iconified_clients)
        x = self.x + ((nb % self.columns) * (self.side + self.space))
        y = self.y + ((nb / self.columns) * ((self.side * 2) + self.space))

        iconified_client = IconifiedClient(self, client, x, y)
        self.iconified_clients.append(iconified_client)

    def remove_iconified_client(self, client):
        iconified_client = self.get_iconified_client_by_client(client)
        if iconified_client is None:
            return

        self.iconified_clients.remove(iconified_client)
        iconified_client.destroy()
        del iconified_client
        self.reorganize()

    def withdraw_iconified_client(self, client):
        # check for false withdraw due to the client hiding
        if client.states.icccm_state == client.withdrawn:
            self.remove_iconified_client(client)

    def reorganize(self):
        pointer = self.wm.root.query_pointer()
        x = self.x
        y = self.y
        for i, iconified_client in enumerate(self.iconified_clients):
            if i % self.columns == 0:
                x = self.x
            else:
                x += self.side + self.space
            if i != 0 and i % self.columns == 0:
                y += (self.side * 2) + self.space
            iconified_client.replace(x, y)
            if (iconified_client.contain(pointer.root_x, pointer.root_y) and
                self.wm.client_at_point(pointer.root_x, pointer.root_y) \
                    is None):
                iconified_client.enter()

    def get_iconified_client_by_client(self, client):
        for iconified_client in self.iconified_clients:
            if iconified_client.client == client:
                return iconified_client
        return None

    def get_iconified_client(self, window):
        for iconified_client in self.iconified_clients:
            if (iconified_client.mini_icon == window or
                iconified_client.title == window):
                return iconified_client
        return None

    def handle_events(self):
        while self.display.pending_events():
            event = self.display.next_event()
            # we are only interested in events with a window attribute
            if not hasattr(event, 'window'):
                continue

            iconified_client = self.get_iconified_client(event.window)
            if iconified_client is None:
                continue

            if event.type == X.Expose:
                iconified_client.update()
            elif event.type == X.ButtonPress:
                iconified_client.uniconify()
            elif event.type == X.EnterNotify:
                iconified_client.enter()
            elif event.type == X.LeaveNotify:
                iconified_client.leave()

    def destroy(self):
        remove_post_hook(Client.iconify, self.add_iconified_client)
        remove_post_hook(Client.uniconify, self.remove_iconified_client)
        remove_post_hook(Client.withdraw, self.withdraw_iconified_client)
        remove_post_hook(Client.destroy, self.remove_iconified_client)
        try:
            self.wm.displays.remove(self.display_wrapper)
        except ValueError:
            print 'pager socket wrapper not in the connection list!'
